@@ -28,6 +28,12 @@
 ;; To enable, add to your ~/.emacs:
 ;; (require 'erc-button)
 ;; (erc-button-mode 1)
+;;
+;; Todo:
+;; * Rewrite all this to do the same, but use button.el from GNU Emacs
+;; if it's available for xemacs too.  Why? button.el is much faster,
+;; and much more elegant, and solves the problem we get with large buffers
+;; and a large erc-button-marker-list.
 
 
 ;;; Code:
@@ -37,20 +43,23 @@
 
 ;;; Minor Mode
 
-;;;###autoload (autoload 'erc-button-mode "erc-button")
+(defgroup erc-button nil
+  "Define how text can be turned into clickable buttons."
+  :group 'erc)
+
+;;;###autoload (autoload 'erc-button-mode "erc-button" nil t)
 (define-erc-module button nil
   "This mode buttonizes all messages according to `erc-button-alist'."
   ((add-hook 'erc-insert-modify-hook 'erc-button-add-buttons 'append)
-   (add-hook 'erc-send-modify-hook 'erc-button-add-buttons 'append))
+   (add-hook 'erc-send-modify-hook 'erc-button-add-buttons 'append)
+   (add-hook 'erc-complete-functions 'erc-button-next))
   ((remove-hook 'erc-insert-modify-hook 'erc-button-add-buttons)
-   (remove-hook 'erc-send-modify-hook 'erc-button-add-buttons)))
-
-(define-key erc-mode-map "\C-c\C-w\C-b" 'widget-backward)
-(define-key erc-mode-map "\C-c\C-w\C-f" 'widget-forward)
+   (remove-hook 'erc-send-modify-hook 'erc-button-add-buttons)
+   (remove-hook 'erc-complete-functions 'erc-button-next)))
 
 ;;; Variables
 
-(defconst erc-button-version "$Revision: 1.32 $"
+(defconst erc-button-version "$Revision: 1.47 $"
   "ERC button mode revision.")
 
 (defcustom erc-button-face 'bold
@@ -69,92 +78,104 @@ above them."
   :type 'face
   :group 'erc-faces)
 
-(defcustom erc-button-url-regexp "\\b\\(\\(www\\.\\|\\(s?https?\\|ftp\\|file\\|gopher\\|news\\|telnet\\|wais\\|mailto\\):\\)\\(//[-a-zA-Z0-9_.]+:[0-9]*\\)?\\([-a-zA-Z0-9_=!?#$@~`%&*+|\\/:;.,]\\|\\w\\)+\\([-a-zA-Z0-9_=#$@~`%&*+|\\/]\\|\\w\\)\\)"
+(defcustom erc-button-url-regexp "\\(www\\.\\|\\(s?https?\\|ftp\\|file\\|gopher\\|news\\|telnet\\|wais\\|mailto\\):\\)\\(//[-a-zA-Z0-9_.]+:[0-9]*\\)?[-a-zA-Z0-9_=!?#$@~`%&*+\\/:;.,]+[-a-zA-Z0-9_=#$@~`%&*+\\/]"
   "Regular expression that matches URLs."
-  :group 'erc
+  :group 'erc-button
   :type 'regexp)
 
+(defcustom erc-button-wrap-long-urls nil
+  "If non-nil, \"long\" URLs matching `erc-button-url-regexp' will be wrapped.
+
+If this variable is a number, consider URLs longer than its value to
+be \"long\".  If t, URLs will be considered \"long\" if they are
+longer than `erc-fill-column'."
+  :group 'erc-button
+  :type '(choice integer boolean))
+
 (defcustom erc-button-buttonize-nicks 100
   "Flag indicating whether nicks are buttonized or not.
 When the number of nicks on the channel is equal or smaller than the value
 of this variable, then nicks are buttonized.  When the number of nicks on
 the channel is larger, nicks are no longer buttonized."
-  :group 'erc
+  :group 'erc-button
   :type 'integer)
 
 (defcustom erc-button-rfc-url "http://www.faqs.org/rfcs/rfc%s.html"
   "*URL used to browse rfc references.
 %s is replaced by the number."
-  :group 'erc
+  :group 'erc-button
   :type 'string)
 
 (defcustom erc-button-google-url "http://www.google.com/search?q=%s"
   "*URL used to browse Google search references.
 %s is replaced by the search string."
-  :group 'erc
+  :group 'erc-button
   :type 'string)
 
 (defcustom erc-button-alist
-  `(("(\\(\\([^~\n \t@][^\n \t@]*\\)@\\([a-zA-Z0-9.:-]+\\)\\)" 1 t finger 2 3)
+  ;; Since the callback is only executed when the user is clicking on
+  ;; a button, it makes no sense to optimize performance by
+  ;; bytecompiling lambdas in this alist.  On the other hand, it makes
+  ;; things hard to maintain.
+  '(("(\\(\\([^~\n \t@][^\n \t@]*\\)@\\([a-zA-Z0-9.:-]+\\)\\)" 1 t finger 2 3)
     ("\\bInfo:[\"]\\([^\"]+\\)[\"]" 0 t Info-goto-node 1)
     ("<URL: *\\([^<> ]+\\) *>" 0 t browse-url 1)
     ("\\b\\(Ward\\|Wiki\\|WardsWiki\\|TheWiki\\):\\([A-Z][a-z]+\\([A-Z][a-z]+\\)+\\)"
-     0 t ,(byte-compile
-	   (lambda (page)
-	     (browse-url (concat "http://c2.com/cgi-bin/wiki?" page))))
+     0 t (lambda (page)
+	   (browse-url (concat "http://c2.com/cgi-bin/wiki?" page)))
      2)
     ("EmacsWiki:\\([A-Z][a-z]+\\([A-Z][a-z]+\\)+\\)" 0 t erc-browse-emacswiki 1)
     (erc-button-url-regexp 0 t browse-url 0)
-    (erc-nick-regexp 0 (or (and (numberp erc-button-buttonize-nicks)
+    (channel-members 0 (or (and (numberp erc-button-buttonize-nicks)
 				(> erc-button-buttonize-nicks
 				   (length channel-members)))
 			   erc-button-buttonize-nicks)
 		     erc-nick-popup 0)
     ("\\bGoogle:\\([^ \t\n\r\f]+\\)"
-     0 t ,(byte-compile
-	   (lambda (keywords)
-	     (browse-url (format erc-button-google-url keywords))))
+     0 t (lambda (keywords)
+	   (browse-url (format erc-button-google-url keywords)))
      1)
     ("\\brfc[#: ]?\\([0-9]+\\)"
-     0 t ,(byte-compile
-	   (lambda (num)
-	     (browse-url (format erc-button-rfc-url num))))
+     0 t (lambda (num)
+	   (browse-url (format erc-button-rfc-url num)))
      1)
     ("\\s-\\(@\\([0-9][0-9][0-9]\\)\\)"
-     1 t ,(byte-compile
-	   (lambda (beats)
-	     (let* ((seconds (- (* (string-to-number beats) 86.4)
-				3600
-				(- (car (current-time-zone)))))
-		    (hours (mod (floor seconds 3600) 24))
-		    (minutes (mod (round seconds 60) 60)))
-	       (message (format "@%s is %d:%02d local time"
-				beats hours minutes)))))
+     1 t (lambda (beats)
+	   (let* ((seconds (- (* (string-to-number beats) 86.4)
+			      3600
+			      (- (car (current-time-zone)))))
+		  (hours (mod (floor seconds 3600) 24))
+		  (minutes (mod (round seconds 60) 60)))
+	     (message (format "@%s is %d:%02d local time"
+			      beats hours minutes))))
      2)
     ("[`]\\([a-zA-Z][-a-zA-Z_0-9]+\\)[']"
-     1 t ,(byte-compile
-	   (lambda (symbol-name)
-	     (let ((symbol (intern-soft symbol-name)))
-	       (cond ((and symbol (fboundp symbol))
-		      (describe-function symbol))
-		     ((and symbol (boundp symbol))
-		      (describe-variable symbol))
-		     (t (apropos symbol-name))))))
+     1 t (lambda (symbol-name)
+	   (let ((symbol (intern-soft symbol-name)))
+	     (cond ((and symbol (fboundp symbol))
+		    (describe-function symbol))
+		   ((and symbol (boundp symbol))
+		    (describe-variable symbol))
+		   (t (apropos symbol-name)))))
      1))
   "*Alist of regexps matching buttons in ERC buffers.
-
 Each entry has the form (REGEXP BUTTON FORM CALLBACK PAR...), where
-REGEXP: is the string matching text around the button
-or a symbol indicating a variable holding that string,
-BUTTON: is the number of the regexp grouping actually matching the button,
-FORM: is a lisp expression which must eval to true for the button to
-be added,
-CALLBACK: is the function to call when the user push this button, and each
-PAR: is a number of a regexp grouping whose text will be passed to CALLBACK.
-
-CALLBACK can also be a variable, in that case the value of that
-variable is the real callback function."
-  :group 'erc
+REGEXP is the string matching text around the button or a symbol
+  indicating a variable holding that string, or a list of strings, or an
+  alist with the strings in the car (such as `channel-members').
+  Note that entries in lists or alists are considered to be nicks or
+  other complete words.  Therefore they are enclosed in \\< and \\>
+  while searching.
+BUTTON is the number of the regexp grouping actually matching the
+  button,
+FORM is a lisp expression which must eval to true for the button to
+  be added,
+CALLBACK is the function to call when the user push this button.
+  CALLBACK can also be a symbol.  Its variable value will be used
+  as the callback function.
+PAR is a number of a regexp grouping whose text will be passed to
+  CALLBACK.  There can be several PAR arguments."
+  :group 'erc-button
   :type '(repeat (list (choice regexp variable)
 		       (integer :tag "Button")
 		       (sexp :tag "Form")
@@ -165,13 +186,19 @@ variable is the real callback function."
 
 (defcustom erc-emacswiki-url "http://www.emacswiki.org/cgi-bin/wiki.pl?"
   "*URL of the EmacsWiki Homepage."
-  :group 'erc
+  :group 'erc-button
   :type 'string)
 
 (defvar erc-button-keymap
-  '(keymap
-    (13 . widget-button-press)
-    (down-mouse-2 . widget-button-click)))
+  (let ((map (make-sparse-keymap)))
+    (define-key map (kbd "RET") 'erc-button-press-button)
+    (if (featurep 'xemacs)
+	(define-key map (kbd "<button2>") 'erc-button-click-button)
+      (define-key map (kbd "<mouse-2>") 'erc-button-click-button))
+    (define-key map (kbd "TAB") 'erc-button-next)
+    (set-keymap-parent map erc-mode-map)
+    map)
+  "Local keymap for ERC buttons.")
 
 (defvar erc-button-syntax-table
   (let ((table (make-syntax-table)))
@@ -184,6 +211,7 @@ variable is the real callback function."
     (modify-syntax-entry ?` "w" table)
     (modify-syntax-entry ?' "w" table)
     (modify-syntax-entry ?^ "w" table)
+    (modify-syntax-entry ?- "w" table)
     (modify-syntax-entry ?_ "w" table)
     (modify-syntax-entry ?| "w" table)
     table)
@@ -191,9 +219,6 @@ variable is the real callback function."
 This syntax table should make all the legal nick characters word
 constituents.")
 
-(defvar erc-button-marker-list nil)
-(make-variable-buffer-local 'erc-button-marker-list)
-
 (defun erc-button-add-buttons ()
   "Find external references in the current buffer and make buttons of them.
 \"External references\" are things like URLs, as
@@ -204,52 +229,74 @@ specified by `erc-button-alist'."
       (let ((buffer-read-only nil)
 	    (inhibit-point-motion-hooks t)
 	    (alist erc-button-alist)
-	    beg entry regexp)
+	    entry regexp data)
 	(erc-button-remove-old-buttons)
-	(setq beg (point-min))
-	(while (setq entry (pop alist))
-	(setq regexp (or (and (stringp (car entry)) (car entry))
-			 (and (boundp (car entry))
-			      (symbol-value (car entry)))))
-	(when (stringp regexp)
-	  (goto-char beg)
-	  (while (re-search-forward regexp nil t)
-	    (let* ((start (and entry (match-beginning (nth 1 entry))))
-		   (end (and entry (match-end (nth 1 entry))))
-		   (from (match-beginning 0)))
-	      (when (and (or (eq t (nth 2 entry))
-			     (eval (nth 2 entry)))
-			 (not (text-property-not-all
-			       start end 'erc-callback nil)))
-		;; That optional form returned non-nil, so we add the
-		;; button.
-		(erc-button-add-button
-		 start end 'erc-button-push
-		 (car (push (set-marker (make-marker) from)
-			    erc-button-marker-list))))))))))))
+	(dolist (entry alist)
+	  (setq regexp (or (and (stringp (car entry)) (car entry))
+			   (and (boundp (car entry))
+				(symbol-value (car entry)))))
+	  (cond ((stringp regexp)
+		 (erc-button-add-buttons-1 regexp entry))
+		((and (listp regexp) (stringp (car regexp)))
+		 (dolist (r regexp)
+		   (erc-button-add-buttons-1
+		    (concat "\\<" (regexp-quote r) "\\>")
+		    entry)))
+		((and (listp regexp) (listp (car regexp))
+		      (stringp (caar regexp)))
+		 (dolist (elem regexp)
+		   (erc-button-add-buttons-1
+		    (concat "\\<" (regexp-quote (car elem)) "\\>")
+		    entry)))))))))
+
+(defun erc-button-add-buttons-1 (regexp entry)
+  "Search through the buffer for matches to ENTRY and add buttons."
+  (goto-char (point-min))
+  (while (re-search-forward regexp nil t)
+    (let ((start (match-beginning (nth 1 entry)))
+	  (end (match-end (nth 1 entry)))
+	  (form (nth 2 entry))
+	  (fun (nth 3 entry))
+	  (data (mapcar 'match-string (nthcdr 4 entry))))
+      (when (or (eq t form)
+		(eval form))
+	(erc-button-add-button start end fun data regexp)))))
 
 (defun erc-button-remove-old-buttons ()
   "Remove all existing buttons.
 This is called with narrowing in effect, just before the text is
-buttonized again.  `erc-button-marker-list' stores the positions
-of all buttons.  Only the visible buttons in that list are removed.
-Removing a button means to make set the `erc-callback' property to
-nil and to remove the position from the marker list."
-  ;; new-list stores the button positions which are not removed
-  (let (marker entry new-list)
-    (while (setq marker (pop erc-button-marker-list))
-      (if (or (< marker (point-min)) (>= marker (point-max)))
-	  (push marker new-list)
-	(goto-char marker)
-	(when (setq entry (erc-button-entry))
-	  (put-text-property (match-beginning (nth 1 entry))
-			     (match-end (nth 1 entry))
-			     'erc-callback nil))
-	(set-marker marker nil)))
-    (setq erc-button-marker-list new-list)))
+buttonized again.  Removing a button means to remove all the properties
+that `erc-button-add-button' adds, except for the face."
+  (remove-text-properties
+   (point-min) (point-max)
+   '(erc-callback nil
+     erc-data nil
+     mouse-face nil
+     local-map nil)))
 
-(defun erc-button-add-button (from to fun &optional data)
+(defun erc-button-add-button (from to fun &optional data regexp)
   "Create a button between FROM and TO with callback FUN and data DATA."
+  ;; Really nasty hack to <URL: > ise urls, and line-wrap them if
+  ;; they're going to be wider than `erc-fill-column'.
+  ;; This could be a lot cleaner, but it works for me -- lawrence.
+  (when (and erc-button-wrap-long-urls
+             (string= regexp erc-button-url-regexp)
+             (> (- to from) (if (numberp erc-button-wrap-long-urls)
+                                erc-button-wrap-long-urls
+                                erc-fill-column)))
+    (let ((erc-fill-column (if (numberp erc-button-wrap-long-urls)
+                               erc-button-wrap-long-urls
+                               erc-fill-column))
+          pos)
+      (setq to (prog1 (point-marker) (insert ">"))
+            from   (prog2 (goto-char from) (point-marker) (insert "<URL: "))
+            pos (copy-marker from))
+      (while (> (- to pos) erc-fill-column)
+        (goto-char (+ pos erc-fill-column))
+        (insert "\n      ")             ; This ought to figure out
+                                        ; what type of filling we're
+                                        ; doing, and indent accordingly.
+        (move-marker pos (point)))))
   (when erc-button-face
     (erc-button-add-face from to erc-button-face))
   (add-text-properties
@@ -258,16 +305,17 @@ nil and to remove the position from the marker list."
 	       (list 'mouse-face erc-button-mouse-face))
 	  (list 'erc-callback fun)
 	  (list 'local-map erc-button-keymap)
+          (list 'rear-nonsticky t)
 	  (and data (list 'erc-data data))))
   (widget-convert-button 'link from to :action 'erc-widget-press-button))
 
 (defun erc-button-add-face (from to face)
   "Add FACE to the region between FROM and TO."
-;; If we just use `add-text-property', then this will overwrite any
-;; face text property already used for the button.  It will not be
-;; merged correctly.  If we use overlays, then redisplay will be very
-;; slow with lots of buttons.  This is why we manually merge face text
-;; properties.
+  ;; If we just use `add-text-property', then this will overwrite any
+  ;; face text property already used for the button.  It will not be
+  ;; merged correctly.  If we use overlays, then redisplay will be
+  ;; very slow with lots of buttons.  This is why we manually merge
+  ;; face text properties.
   (let ((old (erc-list (get-text-property from 'face)))
 	(pos from)
 	(end (next-single-property-change from 'face nil to))
@@ -283,9 +331,12 @@ nil and to remove the position from the marker list."
 	    old (erc-list (get-text-property pos 'face))
 	    end (next-single-property-change pos 'face nil to)))))
 
-(defun erc-widget-press-button (elems el)
-  (goto-char (widget-get elems :from))
-  (erc-button-press-button))
+(defun erc-button-click-button (event)
+  "Call `erc-button-press-button'."
+  (interactive "e")
+  (save-excursion
+    (mouse-set-point event)
+    (erc-button-press-button)))
 
 (defun erc-button-press-button ()
   "Check text at point for a callback function.
@@ -294,45 +345,27 @@ call it with the value of the `erc-data' text property."
   (interactive)
   (let* ((data (get-text-property (point) 'erc-data))
 	 (fun (get-text-property (point) 'erc-callback)))
-    (when fun
-      (funcall fun data))))
+    (unless fun
+      (message "No button at point"))
+    (when (and fun (symbolp fun) (not (fboundp fun)))
+      (error "Function %S is not bound"))
+    (apply fun data)))
 
-(defun erc-button-push (marker)
-  "Push button starting at MARKER."
-  (save-excursion
-    (goto-char marker)
-    (let* ((entry (erc-button-entry))
-	   (inhibit-point-motion-hooks t)
-	   (fun (nth 3 entry))
-	   (args (mapcar (lambda (group)
-			   (let ((string (match-string group)))
-			     (set-text-properties
-			      0 (length string) nil string)
-			     string))
-			 (nthcdr 4 entry))))
-      (cond
-       ((functionp fun)
-	(apply fun args))
-       ((and (boundp fun)
-	     (fboundp (symbol-value fun)))
-	(apply (symbol-value fun) args))
-       (t
-	(message "You must define `%S' to use this button"
-		 (cons fun args)))))))
-
-(defun erc-button-entry ()
-  "Return the first entry in `erc-button-alist' matching this place."
-  (let ((alist erc-button-alist)
-	entry)
-    (with-syntax-table erc-button-syntax-table
-      (while alist
-	(setq entry (pop alist))
-	(if (looking-at (or (and (stringp (car entry)) (car entry))
-			    (and (boundp (car entry))
-				 (symbol-value (car entry)))))
-	    (setq alist nil)
-	  (setq entry nil)))
-      entry)))
+(defun erc-button-next ()
+  "Go to the next button in this buffer."
+  (interactive)
+  (let ((here (point)))
+    (when (< here (erc-beg-of-input-line))
+      (while (and (get-text-property here 'erc-callback)
+                  (not (= here (point-max))))
+        (setq here (1+ here)))
+      (while (and (not (get-text-property here 'erc-callback))
+                  (not (= here (point-max))))
+        (setq here (1+ here)))
+      (if (< here (point-max))
+          (goto-char here)
+        (error "No next button"))
+      t)))
 
 (defun erc-browse-emacswiki (thing)
   "Browse to thing in the emacs-wiki."
@@ -340,21 +373,6 @@ call it with the value of the `erc-data' text property."
 
 ;;; Nickname buttons:
 
-(defvar erc-nick-regexp nil)
-(make-variable-buffer-local 'erc-nick-regexp)
-
-(require 'regexp-opt)
-(defun erc-recompute-nick-regexp ()
-  (if channel-members
-      (setq erc-nick-regexp
-	    (let ((max-specpdl-size 3000))
-	      (concat "\\b\\("
-		      (regexp-opt (mapcar #'car channel-members))
-		      "\\)\\b")))
-    (setq erc-nick-regexp nil)))
-
-(add-hook 'erc-channel-members-changed-hook 'erc-recompute-nick-regexp)
-
 (defcustom erc-nick-popup-alist
   '(("DeOp"  . (erc-cmd-DEOP nick))
     ("Kick"  . (erc-cmd-KICK (concat nick " "
@@ -377,7 +395,7 @@ Examples:
    (format
     \"ldapsearch -x -P 2 -h db.debian.org -b dc=debian,dc=org ircnick=%s\"
     nick)))"
-  :group 'erc
+  :group 'erc-button
   :type '(repeat (cons (string :tag "Op")
 		       sexp)))
 
